@@ -19,6 +19,7 @@ import os
 import subprocess
 import tempfile
 import time
+import traceback
 
 from google.cloud import ndb
 import pygit2
@@ -28,23 +29,16 @@ from . import repos
 from . import models
 from . import vulnerability_pb2
 
-COMMIT_RANGE_LIMIT = 4
-
 TAG_PREFIX = 'refs/tags/'
 
-# Used in cases where an earlier commit in a regression range cannot be
-# determined.
-UNKNOWN_COMMIT = 'unknown'
-
-AffectedResult = collections.namedtuple(
-    'AffectedResult', 'tags commits affected_ranges '
-    'regress_commits fix_commits')
+AffectedResult = collections.namedtuple('AffectedResult',
+                                        'tags commits affected_ranges')
 
 AnalyzeResult = collections.namedtuple('AnalyzeResult', 'has_changes commits')
 
 TagsInfo = collections.namedtuple('TagsInfo', 'tags latest_tag')
 
-_DATASTORE_BATCH_SIZE = 20000
+_DATASTORE_BATCH_SIZE = 10000
 _DATASTORE_BATCH_SLEEP = 5
 
 
@@ -91,96 +85,84 @@ class RepoAnalyzer:
   def __init__(self, detect_cherrypicks=True):
     self.detect_cherrypicks = detect_cherrypicks
 
-  def get_affected(self, repo, regress_commit_or_range, fix_commit_or_range):
+  def get_affected(self,
+                   repo,
+                   regress_commits,
+                   fix_commits,
+                   limit_commits=None):
     """"Get list of affected tags and commits for a bug given regressed and
     fixed commits."""
-    regress_commits = _get_commit_range(repo, regress_commit_or_range)
-    if len(regress_commits) > COMMIT_RANGE_LIMIT:
-      raise ImpactError('Too many commits in regression range.')
-
-    # If multiple, assume the first commit in the regression range cause the
-    # regression.
-    if regress_commits:
-      regress_commit = regress_commits[0]
-    else:
-      regress_commit = None
-
-    fix_commits = _get_commit_range(repo, fix_commit_or_range)
-    if len(fix_commits) > COMMIT_RANGE_LIMIT:
-      logging.warning('Too many commits in fix range.')
-      # Rather than bail out here and potentially leaving a Bug as "unfixed"
-      # indefinitely, we continue.
-
-    # If multiple, assume the last commit is necessary for fixing the
-    # regression.
-    if fix_commits:
-      fix_commit = fix_commits[-1]
-    else:
-      fix_commit = None
-
     affected_commits, affected_ranges, tags = self._get_affected_range(
-        repo, regress_commit, fix_commit)
+        repo, regress_commits, fix_commits, limit_commits=limit_commits)
 
-    if len(regress_commits) > 1 or len(fix_commits) > 1:
-      # Don't return ranges if input regressed and fixed commits are not single
-      # commits.
-      affected_ranges = []
+    return AffectedResult(tags, affected_commits, affected_ranges)
 
-    return AffectedResult(tags, affected_commits, affected_ranges,
-                          regress_commits, fix_commits)
-
-  def _get_affected_range(self, repo, regress_commit, fix_commit):
+  def _get_affected_range(self,
+                          repo,
+                          regress_commits,
+                          fix_commits,
+                          limit_commits=None):
     """Get affected range."""
     range_collector = RangeCollector()
     commits = set()
     seen_commits = set()
     tags = set()
     commits_to_tags = _get_commit_to_tag_mappings(repo)
-
-    is_complete_range = regress_commit and fix_commit
-    # Only actually detect cherrypicks if the input range includes both
-    # "introduced" and "fixed". Otherwise we may get false positives.
-    detect_cherrypicks = self.detect_cherrypicks and is_complete_range
+    branch_to_limit = {}
 
     branches = []
+    detect_cherrypicks = self.detect_cherrypicks and not limit_commits
     if detect_cherrypicks:
       # Check all branches for cherry picked regress/fix commits (sorted for
       # determinism).
       branches = sorted(repo.branches.remote)
     else:
-      if fix_commit:
-        # If a fix commit is available, we don't really need to analyze any
-        # other branches.
-        branches = [repo.head.name.replace('refs/heads/', 'origin/')]
-      elif regress_commit:
+      if limit_commits:
+        for limit_commit in limit_commits:
+          current_branches = _branches_with_commit(repo, limit_commit)
+          for branch in current_branches:
+            branch_to_limit[branch] = limit_commit
+
+          branches.extend(current_branches)
+      elif fix_commits:
+        # TODO(ochang): Remove this check. This behaviour should only be keyed
+        # on `limit_commits`.
+        # If not detecting cherry picks, take only branches that contain the fix
+        # commit. Otherwise we may have false positives.
+        for fix_commit in fix_commits:
+          branches.extend(_branches_with_commit(repo, fix_commit))
+      elif regress_commits:
         # If only a regress commit is available, we need to find all branches
         # that it reaches.
-        branches = _branches_with_commit(repo, regress_commit)
+        for regress_commit in regress_commits:
+          branches.extend(_branches_with_commit(repo, regress_commit))
 
     for branch in branches:
       ref = 'refs/remotes/' + branch
 
       # Get the earliest equivalent commit in the regression range.
-      if detect_cherrypicks:
+      equivalent_regress_commit = None
+      for regress_commit in regress_commits:
         logging.info('Finding equivalent regress commit to %s in %s',
                      regress_commit, ref)
         equivalent_regress_commit = self._get_equivalent_commit(
             repo, ref, regress_commit, detect_cherrypicks=detect_cherrypicks)
-      else:
-        equivalent_regress_commit = regress_commit
+        if equivalent_regress_commit:
+          break
 
       # If regress_commits is provided, then we should find an equivalent.
-      if not equivalent_regress_commit and regress_commit:
+      if not equivalent_regress_commit and regress_commits:
         continue
 
       # Get the latest equivalent commit in the fix range.
-      if detect_cherrypicks:
+      equivalent_fix_commit = None
+      for fix_commit in fix_commits:
         logging.info('Finding equivalent fix commit to %s in %s', fix_commit,
                      ref)
         equivalent_fix_commit = self._get_equivalent_commit(
             repo, ref, fix_commit, detect_cherrypicks=detect_cherrypicks)
-      else:
-        equivalent_fix_commit = fix_commit
+        if equivalent_fix_commit:
+          break
 
       range_collector.add(equivalent_regress_commit, equivalent_fix_commit)
 
@@ -196,13 +178,14 @@ class RepoAnalyzer:
         continue
 
       seen_commits.add((equivalent_regress_commit, end_commit))
-      cur_commits, cur_tags = _get_commit_and_tag_list(
+      cur_commits, cur_tags = get_commit_and_tag_list(
           repo,
           equivalent_regress_commit,
           end_commit,
           commits_to_tags=commits_to_tags,
           include_start=True,
-          include_end=include_end)
+          include_end=include_end,
+          limit_commit=branch_to_limit.get(branch))
       commits.update(cur_commits)
       tags.update(cur_tags)
 
@@ -252,24 +235,6 @@ class RepoAnalyzer:
     return None
 
 
-def _get_commit_range(repo, commit_or_range):
-  """Get a commit range."""
-  if not commit_or_range:
-    return []
-
-  if ':' not in commit_or_range:
-    return [commit_or_range]
-
-  start_commit, end_commit = commit_or_range.split(':')
-  if start_commit == UNKNOWN_COMMIT:
-    # Special case: No information about earlier builds. Assume the end_commit
-    # is the regressing commit as that's the best we can do.
-    return [end_commit]
-
-  commits, _ = _get_commit_and_tag_list(repo, start_commit, end_commit)
-  return commits
-
-
 def _get_commit_to_tag_mappings(repo):
   """Get all commit to tag mappings"""
   mappings = {}
@@ -284,13 +249,21 @@ def _get_commit_to_tag_mappings(repo):
   return mappings
 
 
-def _get_commit_and_tag_list(repo,
-                             start_commit,
-                             end_commit,
-                             commits_to_tags=None,
-                             include_start=False,
-                             include_end=True):
+def get_commit_and_tag_list(repo,
+                            start_commit,
+                            end_commit,
+                            commits_to_tags=None,
+                            include_start=False,
+                            include_end=True,
+                            limit_commit=None):
   """Given a commit range, return the list of commits and tags in the range."""
+  if limit_commit:
+    if str(repo.merge_base(end_commit, limit_commit)) == limit_commit:
+      # Limit commit is an earlier ancestor, so use that as the end of the
+      # range instead.
+      include_end = False
+      end_commit = limit_commit
+
   logging.info('Getting commits %s..%s', start_commit, end_commit)
   try:
     walker = repo.walk(end_commit,
@@ -359,19 +332,14 @@ def _throttled_delete(to_delete):
       time.sleep(_DATASTORE_BATCH_SLEEP)
 
 
-def update_affected_commits(bug_id, commits, project, ecosystem, public):
+def update_affected_commits(bug_id, commits, public):
   """Update affected commits."""
   to_put = []
   to_delete = []
 
   for commit in commits:
     affected_commit = models.AffectedCommit(
-        id=bug_id + '-' + commit,
-        bug_id=bug_id,
-        commit=commit,
-        project=project,
-        ecosystem=ecosystem,
-        public=public)
+        id=bug_id + '-' + commit, bug_id=bug_id, commit=commit, public=public)
 
     to_put.append(affected_commit)
 
@@ -387,7 +355,7 @@ def update_affected_commits(bug_id, commits, project, ecosystem, public):
   _throttled_delete(to_delete)
 
 
-def enumerate_versions(package, ecosystem, affected_ranges):
+def enumerate_versions_pre_0_8(package, ecosystem, affected_ranges):
   """Enumerate versions from SEMVER and ECOSYSTEM input ranges."""
   versions = set()
   for affected_range in affected_ranges:
@@ -407,26 +375,116 @@ def enumerate_versions(package, ecosystem, affected_ranges):
   return versions
 
 
-def analyze(vulnerability,
-            analyze_git=True,
-            checkout_path=None,
-            detect_cherrypicks=True,
-            versions_from_repo=True):
-  """Update and analyze a vulnerability based on its input ranges."""
+def enumerate_versions(package, ecosystem, affected_range):
+  """Enumerate versions from SEMVER and ECOSYSTEM input ranges."""
+  versions = set()
+  sorted_events = []
+  limits = []
+
+  # Remove any magic '0' values.
+  zero_event = None
+  for event in affected_range.events:
+    if event.introduced == '0':
+      zero_event = event
+      continue
+
+    if event.introduced or event.fixed:
+      sorted_events.append(event)
+      continue
+
+    if event.limit:
+      limits.append(event.limit)
+
+  sorted_events.sort(key=lambda e: e.introduced or e.fixed)
+  if zero_event:
+    sorted_events.insert(0, zero_event)
+
+  last_introduced = None
+  for event in sorted_events:
+    if event.introduced and not last_introduced:
+      last_introduced = event.introduced
+
+    if last_introduced and event.fixed:
+      current_versions = ecosystem.enumerate_versions(
+          package, last_introduced, event.fixed, limits=limits)
+      if current_versions:
+        versions.update(current_versions)
+      last_introduced = None
+
+  if last_introduced:
+    current_versions = ecosystem.enumerate_versions(
+        package, last_introduced, None, limits=limits)
+    if current_versions:
+      versions.update(current_versions)
+
+  versions = list(versions)
+  ecosystem.sort_versions(versions)
+  return versions
+
+
+def _analyze_git_ranges(repo_analyzer, checkout_path, affected_range,
+                        new_versions, commits, new_introduced, new_fixed):
+  """Analyze git ranges."""
   package_repo_dir = tempfile.TemporaryDirectory()
-  package_repo_url = None
   package_repo = None
 
-  # Repo -> Git range collectors
-  range_collectors = collections.defaultdict(RangeCollector)
-  new_versions = set()
-  commits = set()
+  with tempfile.TemporaryDirectory() as package_repo_dir:
+    if checkout_path:
+      repo_name = os.path.basename(
+          affected_range.repo.rstrip('/')).rstrip('.git')
+      package_repo = repos.ensure_updated_checkout(
+          affected_range.repo, os.path.join(checkout_path, repo_name))
+    else:
+      package_repo = repos.clone_with_retries(affected_range.repo,
+                                              package_repo_dir)
+
+    all_introduced = []
+    all_fixed = []
+    all_limit = []
+    for event in affected_range.events:
+      if event.introduced and event.introduced != '0':
+        all_introduced.append(event.introduced)
+        continue
+
+      if event.fixed:
+        all_fixed.append(event.fixed)
+        continue
+
+      if event.limit:
+        all_limit.append(event.limit)
+        continue
+
+    try:
+      result = repo_analyzer.get_affected(package_repo, all_introduced,
+                                          all_fixed, all_limit)
+    except ImpactError:
+      logging.warning('Got error while analyzing git range: %s',
+                      traceback.format_exc())
+      return new_versions, commits
+
+    for introduced, fixed in result.affected_ranges:
+      if introduced and introduced not in all_introduced:
+        new_introduced.add(introduced)
+
+      if fixed and fixed not in all_fixed:
+        new_fixed.add(fixed)
+
+    new_versions.update(result.tags)
+    commits.update(result.commits)
+
+  return new_versions, commits
+
+
+def _analyze_git_ranges_pre_0_8(repo_analyzer, checkout_path, vulnerability,
+                                range_collectors, new_versions, commits):
+  """Analyze git ranges (pre 0.8)."""
+  package_repo_dir = tempfile.TemporaryDirectory()
+  package_repo = None
 
-  repo_analyzer = RepoAnalyzer(detect_cherrypicks=detect_cherrypicks)
   try:
+    grouped_git_ranges = {}
     for affected_range in vulnerability.affects.ranges:
-      if (affected_range.type != vulnerability_pb2.AffectedRange.GIT or
-          not analyze_git):
+      if affected_range.type != vulnerability_pb2.AffectedRange.GIT:
         continue
 
       # Convert empty values ('') to None.
@@ -434,44 +492,70 @@ def analyze(vulnerability,
       fixed = affected_range.fixed or None
       range_collectors[affected_range.repo].add(introduced, fixed)
 
-    for affected_range in vulnerability.affects.ranges:
-      # Go through existing provided ranges to find additional ranges (via
-      # cherrypicks and branches).
-      if (affected_range.type != vulnerability_pb2.AffectedRange.GIT or
-          not analyze_git):
+      grouped_git_ranges.setdefault(affected_range.repo, []).append(
+          (introduced, fixed))
+
+    for repo_url, affected_ranges in grouped_git_ranges.items():
+      if checkout_path:
+        repo_name = os.path.basename(repo_url.rstrip('/')).rstrip('.git')
+        package_repo = repos.ensure_updated_checkout(
+            repo_url, os.path.join(checkout_path, repo_name))
+      else:
+        package_repo_dir.cleanup()
+        package_repo_dir = tempfile.TemporaryDirectory()
+        package_repo = repos.clone_with_retries(repo_url, package_repo_dir.name)
+
+      all_introduced = []
+      all_fixed = []
+      for (introduced, fixed) in affected_ranges:
+        if introduced:
+          all_introduced.append(introduced)
+
+        if fixed:
+          all_fixed.append(fixed)
+
+      try:
+        result = repo_analyzer.get_affected(package_repo, all_introduced,
+                                            all_fixed)
+      except ImpactError:
+        logging.warning('Got error while analyzing git range: %s',
+                        traceback.format_exc())
         continue
 
-      current_repo_url = affected_range.repo
-      if checkout_path:
-        repo_name = os.path.basename(
-            current_repo_url.rstrip('/')).rstrip('.git')
-        package_repo = repos.ensure_updated_checkout(
-            current_repo_url, os.path.join(checkout_path, repo_name))
-      else:
-        if current_repo_url != package_repo_url:
-          # Different repo from previous one.
-          package_repo_dir.cleanup()
-          package_repo_dir = tempfile.TemporaryDirectory()
-          package_repo_url = current_repo_url
-          package_repo = repos.clone_with_retries(package_repo_url,
-                                                  package_repo_dir.name)
-
-      result = repo_analyzer.get_affected(package_repo,
-                                          affected_range.introduced,
-                                          affected_range.fixed)
       for introduced, fixed in result.affected_ranges:
-        range_collectors[current_repo_url].add(introduced, fixed)
+        range_collectors[repo_url].add(introduced, fixed)
 
       new_versions.update(result.tags)
       commits.update(result.commits)
   finally:
     package_repo_dir.cleanup()
 
+  return range_collectors, new_versions, commits
+
+
+def _analyze_pre_0_8(vulnerability,
+                     analyze_git=True,
+                     checkout_path=None,
+                     detect_cherrypicks=True,
+                     versions_from_repo=True):
+  """Update and analyze a vulnerability based on its input ranges (pre 0.8)."""
+  # Repo -> Git range collectors
+  range_collectors = collections.defaultdict(RangeCollector)
+  new_versions = set()
+  commits = set()
+
+  # Analyze git ranges.
+  if analyze_git:
+    repo_analyzer = RepoAnalyzer(detect_cherrypicks=detect_cherrypicks)
+    _analyze_git_ranges_pre_0_8(repo_analyzer, checkout_path, vulnerability,
+                                range_collectors, new_versions, commits)
+
   # Enumerate ECOSYSTEM and SEMVER ranges.
   ecosystem_helpers = ecosystems.get(vulnerability.package.ecosystem)
   if ecosystem_helpers:
-    versions = enumerate_versions(vulnerability.package.name, ecosystem_helpers,
-                                  vulnerability.affects.ranges)
+    versions = enumerate_versions_pre_0_8(vulnerability.package.name,
+                                          ecosystem_helpers,
+                                          vulnerability.affects.ranges)
   else:
     logging.warning('No ecosystem helpers implemented for %s',
                     vulnerability.package.ecosystem)
@@ -511,3 +595,70 @@ def analyze(vulnerability,
 
   vulnerability.modified.FromDatetime(models.utcnow())
   return AnalyzeResult(True, commits)
+
+
+def analyze(vulnerability,
+            analyze_git=True,
+            checkout_path=None,
+            detect_cherrypicks=True,
+            versions_from_repo=True):
+  """Update and analyze a vulnerability based on its input ranges."""
+  if not vulnerability.affected:
+    return _analyze_pre_0_8(vulnerability, analyze_git, checkout_path,
+                            detect_cherrypicks, versions_from_repo)
+
+  commits = set()
+  has_changes = False
+  for affected_package in vulnerability.affected:
+    versions = []
+    for affected_range in affected_package.ranges:
+      if affected_range.type in (vulnerability_pb2.Range.ECOSYSTEM,
+                                 vulnerability_pb2.Range.SEMVER):
+        # Enumerate ECOSYSTEM and SEMVER ranges.
+        ecosystem_helpers = ecosystems.get(affected_package.package.ecosystem)
+        if ecosystem_helpers:
+          versions.extend(
+              enumerate_versions(affected_package.package.name,
+                                 ecosystem_helpers, affected_range))
+        else:
+          logging.warning('No ecosystem helpers implemented for %s',
+                          affected_package.package.ecosystem)
+
+      new_git_versions = set()
+      new_introduced = set()
+      new_fixed = set()
+
+      # Analyze git ranges.
+      if (analyze_git and
+          affected_range.type == vulnerability_pb2.Range.Type.GIT):
+        repo_analyzer = RepoAnalyzer(detect_cherrypicks=detect_cherrypicks)
+        _analyze_git_ranges(repo_analyzer, checkout_path, affected_range,
+                            new_git_versions, commits, new_introduced,
+                            new_fixed)
+
+      # Add additional versions derived from commits and tags.
+      if versions_from_repo:
+        versions.extend(new_git_versions)
+
+      # Apply changes.
+      for introduced in new_introduced:
+        if (not any(
+            event.introduced == introduced for event in affected_range.events)):
+          has_changes = True
+          affected_range.events.add(introduced=introduced)
+
+      for fixed in new_fixed:
+        if not any(event.fixed == fixed for event in affected_range.events):
+          has_changes = True
+          affected_range.events.add(fixed=fixed)
+
+    for version in sorted(versions):
+      if version not in affected_package.versions:
+        has_changes = True
+        affected_package.versions.append(version)
+
+  if not has_changes:
+    return AnalyzeResult(False, commits)
+
+  vulnerability.modified.FromDatetime(models.utcnow())
+  return AnalyzeResult(True, commits)
